/* $Id: cpl_error.c,v 1.35 2006/07/10 07:16:43 cizzo Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2005 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: cizzo $
 * $Date: 2006/07/10 07:16:43 $
 * $Revision: 1.35 $
 * $Name: cpl-3_1 $
 */


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include <cxmessages.h>

#include "cpl_error_impl.h"

/**
 * @defgroup cpl_error Error handling
 *
 * This module provides functions to maintain the @em cpl_error_code
 * set by any CPL function, similarly to what is done with the @em errno 
 * variable of the standard C library. The following guidelines are 
 * respected:
 *
 * - If no error occurs in a CPL function the @em cpl_error_code will 
 *   remain unchanged.
 * - An error condition in a CPL function will cause the previous
 *   @em cpl_error_code to be overwritten by the new one.
 *
 * A @em cpl_error_code equal to the enumeration constant 
 * @c CPL_ERROR_NONE would indicate no error condition. Note, 
 * however, that the @em cpl_error_code is only set when an error 
 * occurs, and it is not reset by successful function calls.
 * For this reason it may be appropriate in some cases to reset 
 * the @em cpl_error_code using the function @c cpl_error_reset(). 
 * The @em cpl_error_code set by a CPL function can be obtained by 
 * calling the function @c cpl_error_get_code(), but functions of 
 * type @em cpl_error_code would not only return this code directly, 
 * but would also return @c CPL_ERROR_NONE in case of success. Other 
 * CPL functions return zero on success, or a non-zero value to indicate 
 * a change of the @em cpl_error_code, while CPL functions returning 
 * a pointer would flag an error by returning a @c NULL.
 *
 * To each @em cpl_error_code is associated a standard error message,
 * that can be obtained by calling the function @c cpl_error_get_message(). 
 * Conventionally, no CPL function will ever display any error message, 
 * leaving to the caller the decision of how to handle a given error 
 * condition. A call to the function @c cpl_error_get_function() would 
 * return the name of the function where the error occurred, and the 
 * functions @c cpl_error_get_file() and @c cpl_error_get_line() would 
 * also return the name of the source file containing the function 
 * code, and the line number where the error occurred. The function
 * @c cpl_error_get_where() would gather all this items together, in
 * a colon-separated string.
 *
 * @par Synopsis:
 * @code
 *   #include <cpl_error.h>
 * @endcode
 */

/**@{*/


#ifndef CPL_LINESZ
#ifdef FITS_LINESZ
#define CPL_LINESZ FITS_LINESZ
#else
#define CPL_LINESZ 6
#endif
#endif

#ifdef PATH_MAX
#define MAX_FILE_LENGTH PATH_MAX
#else
#define MAX_FILE_LENGTH 4096
#endif
#define MAX_NAME_LENGTH 50
#define MAX_WHERE_LENGTH (2+(MAX_NAME_LENGTH)+(MAX_FILE_LENGTH)+(CPL_LINESZ))

static char cpl_error_where_string[MAX_WHERE_LENGTH] = "";

struct _cpl_error_ {
    cpl_error_code  code;
    char            function[MAX_NAME_LENGTH+1];
    char            file[MAX_FILE_LENGTH+1];
    unsigned        line;
} ;

static cpl_error errors[] = {
    { CPL_ERROR_NONE, "", "", 0 },
    { CPL_ERROR_NONE, "", "", 0 }
};

static cpl_error *const error = errors;
static int error_saved = 0;


/**
 * @brief
 *   Reset the @em cpl_error_code.
 *
 * @return Nothing.
 *
 * This function initialises the @em cpl_error_code to @c CPL_ERROR_NONE.
 */

void cpl_error_reset(void)
{
    error->code = CPL_ERROR_NONE;
    error->function[0] = '\0';
    error->file[0] = '\0';
    error->line = 0;
}


/**
 * @internal
 * @brief
 *   Set the @em cpl_error_code to the specified value.
 *
 * @param code  A @em cpl_error_code
 *
 * @return The specified @em cpl_error_code.
 *
 * Set the @em cpl_error_code to the specified code.
 */

cpl_error_code cpl_error_set_code(cpl_error_code code)
{
    error->code = code;
    return code;
}


/*
 * @brief
 *   Set function name, source file & line number where the CPL error occurred.
 *
 * @param function  Character string with function name
 * @param file      Character string with source file name (__FILE__)
 * @param line      Integer with line number (__LINE__)
 *
 * @return Nothing.
 *
 * Set function name, source file & line number where the CPL error occurred.
 * This function is only provided for the macros
 * cpl_error_set_where(const char * function) &&
 * cpl_error_code cpl_error_set(const char * function, cpl_error_code code)
 */

void cpl_error_set_where_macro(const char *function, const char *file,
                               unsigned line)
{
    strncpy(error->function, function, MAX_NAME_LENGTH);
    strncpy(error->file, file, MAX_FILE_LENGTH);
    error->line = line;
}


/*
 * @brief
 *   Set CPL error code, function name, source file & line number where it
 *   occurred.
 * 
 * @param function Character string with function name
 * @param code     Error code
 * @param file     Character string with source file name (__FILE__)
 * @param line     Integer with line number (__LINE__)
 *
 * @return The specified error code.
 *
 * Set CPL error code, function name, source file & line number where it
 * occurred.
 *
 * This function is only provided for the macro
 * cpl_error_code cpl_error_set(const char * function, cpl_error_code code)
 */

cpl_error_code cpl_error_set_macro(const char * function, cpl_error_code code,
                                   const char * file, unsigned line)
{
    cpl_error_set_code(code);
    cpl_error_set_where_macro(function, file, line);
    return cpl_error_get_code();
}


/**
 * @brief
 *   Get the last @em cpl_error_code set.
 *
 * @return  @em cpl_error_code of last occurred CPL error.
 *
 * Get @em cpl_error_code of last occurred error.
 */

cpl_error_code cpl_error_get_code(void)
{
    return error->code;
}


/**
 * @brief
 *   Get standard message of last occurred CPL error.
 *
 * @return Standard message of last occurred CPL error.
 *
 * Get current standard error message. If the @em cpl_error_code
 * is equal to @c CPL_ERROR_NONE, an empty string is returned.
 */

const char *cpl_error_get_message(void)
{
    const char  *   message ;
    switch (error->code) {
        case CPL_ERROR_NONE: 
            message = "" ; 
            break ;
        case CPL_ERROR_DUPLICATING_STREAM: 
            message = "Cannot duplicate output stream" ; 
            break;
        case CPL_ERROR_ASSIGNING_STREAM: 
            message = "Cannot associate a stream with a file descriptor" ; 
            break ;
        case CPL_ERROR_FILE_IO: 
            message = "File access permission denied" ; 
            break ;
        case CPL_ERROR_BAD_FILE_FORMAT: 
            message = "Bad file format" ; 
            break ;
        case CPL_ERROR_FILE_ALREADY_OPEN: 
            message = "File already open" ; 
            break ;
        case CPL_ERROR_FILE_NOT_CREATED: 
            message = "File cannot be created" ; 
            break ;
        case CPL_ERROR_FILE_NOT_FOUND: 
            message = "File not found" ; 
            break ;
        case CPL_ERROR_DATA_NOT_FOUND: 
            message = "Data not found" ; 
            break ;
        case CPL_ERROR_ACCESS_OUT_OF_RANGE: 
            message = "Access beyond boundaries" ; 
            break ;
        case CPL_ERROR_NULL_INPUT: 
            message = "Null input data" ; 
            break ;
        case CPL_ERROR_INCOMPATIBLE_INPUT: 
            message = "Input data do not match" ;
            break ;
        case CPL_ERROR_ILLEGAL_INPUT: 
            message = "Illegal input" ; 
            break ;
        case CPL_ERROR_ILLEGAL_OUTPUT: 
            message = "Illegal output" ; 
            break ;
        case CPL_ERROR_UNSUPPORTED_MODE: 
            message = "Unsupported mode" ; 
            break ;
        case CPL_ERROR_SINGULAR_MATRIX: 
            message = "Singular matrix" ; 
            break ;
        case CPL_ERROR_DIVISION_BY_ZERO: 
            message = "Division by zero" ; 
            break ;
        case CPL_ERROR_TYPE_MISMATCH: 
            message = "Type mismatch" ; 
            break ;
        case CPL_ERROR_INVALID_TYPE: 
            message = "Invalid type" ; 
            break ;
        case CPL_ERROR_CONTINUE: 
            message = "The iterative process did not converge" ; 
            break ;
        case CPL_ERROR_UNSPECIFIED: 
            message = "An unspecified error" ; 
            break ;
        default: 
            message = "" ; 
            break ;
    }
    return message ;
}


/**
 * @brief
 *   Get the function name where the last CPL error occurred.
 *
 * @return Identifier string of the function name where the last CPL error
 *         occurred.
 *
 * Get the function name where the last CPL error occurred.
 */

const char *cpl_error_get_function(void)
{
    return error->function;
}


/**
 * @brief
 *   Get function name, source file and line number where the last 
 *   CPL error occurred.
 *
 * @return String containing function name, source file and line number
 *   separated by colons (:).
 *
 * Get where the last CPL error occurred in the form 
 * @c function_name:source_file:line_number
 */

const char *cpl_error_get_where(void)
{

    (void)snprintf(cpl_error_where_string, MAX_WHERE_LENGTH, "%s:%s:%u",
                   error->function, error->file, error->line);

    return cpl_error_where_string;

}


/**
 * @brief
 *   Get the source code file name where the last CPL error occurred.
 *
 * @return Name of source file name where the last CPL error occurred.
 *
 * Get the source code file name where the last CPL error occurred.
 * If the macro @c __FILE__ is undefined an empty string is returned.
 */

const char *cpl_error_get_file(void)
{
    return error->file; 
}


/**
 * @brief
 *   Get the line number where the last CPL error occurred.
 *
 * @return  Line number of the source file where the last CPL error occurred.
 *
 * Get the line number of the source file where the last CPL error occurred.
 * If the macro @c __LINE__ is undefined 0 is returned.
 */

unsigned cpl_error_get_line(void)
{
    return error->line; 
}


/**
 * @internal
 * @brief 
 *   Push an error onto the stack.
 *
 * @return Nothing.
 *
 * The function saves a possibly set error on the error stack. There is
 * only one stack level, so that calling this function multiple times will
 * overwrite an error which was already saved on the stack before.
 *
 * @see cpl_error_pop()
 */

void
cpl_error_push(void)
{

    cx_assert(error == &errors[0]);

    errors[1].code = error->code;
    errors[1].line = error->line;

    strncpy(errors[1].function, error->function, MAX_NAME_LENGTH);
    errors[1].function[MAX_NAME_LENGTH] = '\0';

    strncpy(errors[1].file, error->file, MAX_FILE_LENGTH);
    errors[1].file[MAX_FILE_LENGTH] = '\0';

    error_saved = 1;

    cpl_error_reset();

    return;

}


/**
 * @internal
 * @brief 
 *   Pop an error from the stack.
 *
 * @return Nothing.
 *
 * The function restores an error which was saved on the stack before. Calling
 * the function a second time without a preceding call to @b cpl_error_push()
 * does not have any effect.
 *
 * @see cpl_error_pop()
 */

void
cpl_error_pop(void)
{

    cx_assert(error == &errors[0]);

    if (error_saved) {
        error->code = errors[1].code;
        error->line = errors[1].line;

        strncpy(error->function, errors[1].function, MAX_NAME_LENGTH);
        error->function[MAX_NAME_LENGTH] = '\0';

        strncpy(error->file, errors[1].file, MAX_FILE_LENGTH);
        error->file[MAX_FILE_LENGTH] = '\0';

        error_saved = 0;

    }

    return;

}
/**@}*/
