/* $Id: cpl_memory.c,v 1.15 2006/07/17 15:34:03 yjung Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2005 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: yjung $
 * $Date: 2006/07/17 15:34:03 $
 * $Revision: 1.15 $
 * $Name: cpl-3_1 $
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <cxmemory.h>
#include <cxmessages.h>
#include <cxstrutils.h>

#include <qfits.h>

#include "cpl_memory.h"


/**
 * @defgroup cpl_memory Memory Management Utilities
 *
 * This module provides the CPL memory management utilities. CPL uses its own
 * memory management system in order to provide almost infinite virtual memory
 * (only limited by the file system size). However, this functionality has
 * certain restrictions. Memory which was allocated using one of the CPL
 * allocator functions must be released with the CPL deallocator function.
 * The same applies to memory blocks allocated and returned by CPL functions.
 * The CPL memory management subsystem uses the services provided by
 * qfits.
 */

/**@{*/

static cxptr _cpl_malloc(cxsize);
static cxptr _cpl_calloc(cxsize, cxsize);
static cxptr _cpl_realloc(cxptr, cxsize);
static void _cpl_free(cxptr);


static cxbool _cpl_memory_is_initialized = FALSE;

static cx_memory_vtable _cpl_memory_services = {
    _cpl_malloc,
    _cpl_calloc,
    _cpl_realloc,
    _cpl_free
};


static cxptr
_cpl_malloc(cxsize nbytes)
{

    return (cxptr)qfits_malloc((size_t) nbytes);

}


static cxptr
_cpl_calloc(cxsize natoms, cxsize nbytes)
{

    return (cxptr)qfits_calloc((size_t)natoms, (size_t) nbytes);

}


static cxptr
_cpl_realloc(cxptr memblk, cxsize nbytes)
{

    return (cxptr)qfits_realloc(memblk, (size_t) nbytes);

}


static void
_cpl_free(cxptr memblk)
{

    qfits_free(memblk);
    return;

}


/**
 * @internal
 * @brief
 *   Initialise the memory management subsystem.
 *
 * @return Nothing.
 *
 * The function initialises the CPL memory management subsystem, i.e. it
 * redirects any request for allocating or deallocating memory to the
 * @c xmemory module of qfits.
 *
 * This must be called, once and only once, before any memory is allocated,
 * either by calling any of the allocator functions directly, or indirecly
 * by calling a CPL function which allocates memory.
 *
 * @note 
 *   Usually, there is no need to call this function directly, since it is
 *   automatically called by the library initialisation function @b cpl_init().
 *
 * @see cpl_init()
 */

void
cpl_memory_init(void)
{

    cx_memory_vtable_set(&_cpl_memory_services);
    _cpl_memory_is_initialized = cx_memory_is_system_malloc() == FALSE;

    return;

}


/**
 * @brief
 *   Allocate @em nbytes bytes.
 *
 * @param nbytes  Number of bytes.
 *
 * @return Pointer to the allocated memory block.
 *
 * The function allocates @em nbytes bytes of memory. The allocated memory
 * is not cleared.
 *
 * @note
 *   If the memory subsytem has not been initialised before calling
 *   this function, the program execution is stopped printing a message to
 *   the error channel showing the current code position.
 *
 */

void *
cpl_malloc(size_t nbytes)
{

    if (_cpl_memory_is_initialized == FALSE) {
        cx_log(CX_LOG_DOMAIN, CX_LOG_LEVEL_ERROR, "%s: CPL memory management "
               "subsystem is not initialized!", CX_CODE_POS);
    }

    return cx_malloc((cxsize)nbytes);

}

        
/**
 * @brief
 *   Allocate memory for @em natoms elements of size @em size.
 *
 * @param natoms  Number of atomic elements.
 * @param nbytes  Element size in bytes.
 *
 * @return Pointer to the allocated memory block.
 *
 * The function allocates memory suitable for storage of @em natoms
 * elements of size @em nbytes bytes. The allocated memory is cleared,
 * i.e. the value 0 is written to each single byte.
 *
 * @note
 *   If the memory subsytem has not been initialised before calling
 *   this function, the program execution is stopped printing a message to
 *   the error channel showing the current code position.
 */

void *
cpl_calloc(size_t natoms, size_t nbytes)
{

    if (_cpl_memory_is_initialized == FALSE) {
        cx_log(CX_LOG_DOMAIN, CX_LOG_LEVEL_ERROR, "%s: CPL memory management "
               "subsystem is not initialized!", CX_CODE_POS);
    }

    return cx_calloc((cxsize)natoms, (cxsize)nbytes);

}


/**
 * @brief
 *   Change the size of a memory block.
 *
 * @param memblk  Pointer to the memory to re-allocate.
 * @param nbytes  New memory block size in bytes.
 *
 * @return Pointer to the allocated memory block.
 *
 * The function changes the size of an already allocated memory block
 * @em memblk to the new size @em nbytes bytes. The contents is unchanged
 * to the minimum of old and new size; newly allocated memory is not
 * initialized. If @em memblk is @c NULL the call to @b cpl_realloc()
 * is equivalent to @b cpl_malloc(), and if @em nbytes is 0 the call
 * is equivalent to @b cpl_free(). Unless @em memblk is @c NULL, it must
 * have been returned by a previous call to @b cpl_malloc(),
 * @b cpl_calloc(), or @b cpl_realloc().
 *
 * @note
 *   - The returned memory block returned on successfull allocation may
 *     not be the same as the one pointed to by @em memblk. Existing
 *     references pointing to locations within the original memory block
 *     might be invalidated!
 *   - If the memory subsytem has not been initialised before calling
 *     this function, the program execution is stopped printing a message to
 *     the error channel showing the current code position.
 *
 * @see cpl_malloc(), cpl_calloc()
 */

void *
cpl_realloc(void *memblk, size_t nbytes)
{

    if (_cpl_memory_is_initialized == FALSE) {
        cx_log(CX_LOG_DOMAIN, CX_LOG_LEVEL_ERROR, "%s: CPL memory management "
               "subsystem is not initialized!", CX_CODE_POS);
    }

    return cx_realloc(memblk, (cxsize)nbytes);

}


/**
 * @brief
 *   Memory block deallocation.
 *
 * @return Nothing.
 *
 * Deallocates a memory block previously allocated by any of the CPL
 * allocator functions.
 *
 * @see cpl_malloc(), cpl_calloc()
 */

void
cpl_free(void *memblk)
{

    if (_cpl_memory_is_initialized == FALSE) {
        cx_log(CX_LOG_DOMAIN, CX_LOG_LEVEL_ERROR, "%s: CPL memory management "
               "subsystem is not initialized!", CX_CODE_POS);
    }

    cx_free(memblk);
    return;

}

/**
 * @brief
 *   Duplicate a string.
 *
 * @param string  String to be duplicated.
 *
 * @return Newly allocated copy of the original string.
 *
 * Duplicates the input string @em string. The newly allocated copy returned
 * to the caller can be deallocated using @b cpl_free().
 *
 * @see cpl_free()
 */

char *cpl_strdup(const char *string)
{

    if (_cpl_memory_is_initialized == FALSE) {
        cx_log(CX_LOG_DOMAIN, CX_LOG_LEVEL_ERROR, "%s: CPL memory management "
               "subsystem is not initialized!", CX_CODE_POS);
    }

    return cx_strdup(string);

}

/**
 * @brief   Tell if there is some memory allocated
 * @return  -1 if the model is off, 1 if it is empty, 0 otherwise
 * The QFITS internal cache is purged before the check. This cache is
 * not something that the user of this function wants to know about.
 */
int cpl_memory_is_empty(void)
{
    qfits_cache_purge() ;
    return qfits_memory_is_empty() ;
}

/**
 * @brief   Display the memory status
 */
void cpl_memory_dump(void)
{
    qfits_cache_purge() ;
    qfits_memory_status() ;
    return ;
}

/**@}*/
