/* $Id: cpl_tools.h,v 1.46 2007/01/17 13:30:06 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2007/01/17 13:30:06 $
 * $Revision: 1.46 $
 * $Name: cpl-3_1 $
 */

#ifndef CPL_TOOLS_H
#define CPL_TOOLS_H

/*-----------------------------------------------------------------------------
   								Includes
 -----------------------------------------------------------------------------*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>

#include <qfits.h>

#include <cxtypes.h>

#include "cpl_propertylist.h"
#include "cpl_error.h"
#include "cpl_vector.h"
#include "cpl_matrix.h"
#include "cpl_msg.h"

/*-----------------------------------------------------------------------------
   								Define
 -----------------------------------------------------------------------------*/

typedef unsigned long long int cpl_flops;

enum _cpl_clock_ {
    CPL_CLOCK_RESET,
    CPL_CLOCK_START,
    CPL_CLOCK_STOP,
    CPL_CLOCK_TIME
};

typedef enum _cpl_clock_ cpl_clock;

#if defined HAVE_LONG_DOUBLE && HAVE_LONG_DOUBLE == 1
typedef long double cpl_long_double;
#define CPL_LDBL_EPSILON LDBL_EPSILON
#define CPL_LDBL_EPSILON_STR "LDBL_EPSILON"
#else
typedef double cpl_long_double;
#define CPL_LDBL_EPSILON DBL_EPSILON
#define CPL_LDBL_EPSILON_STR "DBL_EPSILON"
#endif

/* cpl_test() may be used in the test modules of CPL
   to verify the success of a boolean expression.
   It returns 0 when the test succeeds (evaluates to zero) and 1 when not.

   cpl_test() may be used in like this:

   int nfail = 0;

   nfail += cpl_test( myfunc1() );
   nfail += cpl_test( myfunc2() );

   ....  Additional, similar tests ...

    nfail += cpl_test( cpl_error_get_code() == CPL_ERROR_NONE );

    if (nfail) cpl_msg_error("", "%d test(s) failed", nfail);

    cpl_end();
    return nfail;

 */
#define cpl_test(bool) \
  ((bool) ? (cpl_msg_debug(__FILE__, \
     "OK in line %d (CPL-error state: '%s' in %s): %s",__LINE__, \
       cpl_error_get_message(), cpl_error_get_where(), #bool), 0) \
    : (cpl_msg_error(__FILE__, \
     "Failure in line %d (CPL-error state: '%s' in %s): %s",__LINE__, \
       cpl_error_get_message(), cpl_error_get_where(), #bool), 1))

/* cpl_test_leq() may be used in the test modules of CPL
   to verify the accuracy of a numerical computation.
   It returns 0 when the test succeeds (value <= tolerance) and 1 when not. */
#define cpl_test_leq(value, tolerance) \
  ((value) <= cpl_error_margin*(tolerance) ? (cpl_msg_debug(__FILE__, \
     "OK in line %d: %g <= %g*%g",__LINE__, (double)(value), \
     cpl_error_margin, (double)(tolerance)), 0) : (cpl_msg_error(__FILE__, \
     "Error in line %d: %g > %g*%g",__LINE__, (double)(value), \
     cpl_error_margin, (double)(tolerance)), 1))

/* Acceptable error margin */
#ifndef cpl_error_margin
#define cpl_error_margin 2.0
#endif

/* In the testing modules this can be used to assert success of CPL calls
    that must succeed in order for the remaining tests to be useful  */
#define cpl_assert(bool) do {                                                  \
    /* Evaluate bool just once */                                              \
    const int ok = (bool) ? CPL_TRUE : CPL_FALSE;                              \
    if (ok == CPL_FALSE) {                                                     \
        cpl_msg_error(__FILE__,                                                \
                      "Failure in line %d (CPL-error state: '%s' in %s): %s",  \
                      __LINE__, cpl_error_get_message(),                       \
                      cpl_error_get_where(), #bool);                           \
        cpl_end();                                                             \
        return 1;                                                              \
    }                                                                          \
    if (cpl_error_get_code() != CPL_ERROR_NONE)                                \
        cpl_msg_debug(__FILE__,                                                \
                      "OK in line %d (CPL-error state: '%s' in %s): %s",       \
                      __LINE__, cpl_error_get_message(),                       \
                      cpl_error_get_where(), #bool);                           \
    else                                                                       \
        cpl_msg_debug(__FILE__, "OK in line %d: %s", __LINE__, #bool);         \
  } while (0)


#define cpl_vsprintf(FORMAT, ARGLIST) \
   cpl_vsprintf_macro(FORMAT, ARGLIST, ARGLIST)

/*-----------------------------------------------------------------------------
                            Function prototypes
 -----------------------------------------------------------------------------*/

/* Operations on arrays */
double cpl_tools_get_kth_double(double *, int, int) ;
float cpl_tools_get_kth_float(float *, int, int) ;
int cpl_tools_get_kth_int(int *, int, int) ;

double cpl_tools_get_variance_double(const double *, int, double);
double cpl_tools_get_median_double(double *, int) ;
float cpl_tools_get_median_float(float *, int) ;
int cpl_tools_get_median_int(int *, int) ;

cpl_error_code cpl_tools_sort_double(double *, int) ;
cpl_error_code cpl_tools_sort_int(int *, int) ;

/* Benchmarking functionality */
double cpl_tools_get_cputime(cpl_clock);
void cpl_tools_reset_flops(void);
void cpl_tools_add_flops(cpl_flops);
cpl_flops cpl_tools_get_flops(void);

/* Others */
int cpl_tools_is_power_of_2(int) ;
double cpl_tools_ipow(double, int);
double * cpl_tools_fill_kernel_profile(cpl_kernel, int *) ;
qfits_header * cpl_tools_gen_valid_header(const qfits_header *, 
                                          const cpl_propertylist *, int);

cpl_error_code cpl_matrix_solve_spd(cpl_matrix *, cpl_matrix *);
cpl_matrix * cpl_matrix_product_normal_create(const cpl_matrix *);

char * cpl_vsprintf_macro(const char *, va_list, va_list)
#ifdef __GNUC__
     __attribute__((format (printf, 1, 0)))
#endif
     ;

char * cpl_sprintf(const char *, ...)
#ifdef __GNUC__
     __attribute__((format (printf, 1, 2)))
#endif
     ;

#endif 
