/* $Id: cpl_dfs.c,v 1.34 2007/01/18 07:57:58 llundin Exp $
 *
 * This file is part of the IIINSTRUMENT Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2007/01/18 07:57:58 $
 * $Revision: 1.34 $
 * $Name: cpl-3_1 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string.h>
#include <stdio.h>
#include <cpl_propertylist.h>
#include <cpl_msg.h>
#include <cpl_memory.h>
#include "cpl_dfs.h"
#include "cpl_tools.h"
#include "qfits.h"

/**
 * @defgroup cpl_dfs       DFS related functions
 *
 */

/**@{*/

/*
 * Product keywords aliases
 */

#define PIPEFILE                    "PIPEFILE"
#define PRO_DID                     "ESO PRO DID"
#define PRO_CATG                    "ESO PRO CATG"
#define PRO_TYPE                    "ESO PRO TYPE"
#define PRO_DATE                    "ESO PRO DATE"
#define PRO_DATANCOM                "ESO PRO DATANCOM"
#define PRO_REC_ID                  "ESO PRO REC1 ID"
#define PRO_REC_DRS_ID              "ESO PRO REC1 DRS ID"
#define PRO_REC_PIPE_ID             "ESO PRO REC1 PIPE ID"
#define PRO_REC_RAWi_NAME           "ESO PRO REC1 RAW%d NAME"
#define PRO_REC_RAWi_CATG           "ESO PRO REC1 RAW%d CATG"
#define PRO_REC_CALi_NAME           "ESO PRO REC1 CAL%d NAME"
#define PRO_REC_CALi_CATG           "ESO PRO REC1 CAL%d CATG"
#define PRO_REC_CALi_DATAMD5        "ESO PRO REC1 CAL%d DATAMD5"
#define PRO_REC_PARAMi_NAME         "ESO PRO REC1 PARAM%d NAME"
#define PRO_REC_PARAMi_VALUE        "ESO PRO REC1 PARAM%d VALUE"

#define MAX_PLENGTH (64)

/**
 * @brief    Add product keywords to a pipeline product property list.
 *
 * @param    header          Property list where keywords must be written
 * @param    product_frame   Frame describing the product
 * @param    framelist       List of frames including all raw input frames
 * @param    parlist         Recipe parameter list
 * @param    recid           Recipe name
 * @param    pipeline_id     Pipeline unique identifier
 * @param    dictionary_id   PRO dictionary identifier
 *
 * @return   @c CPL_ERROR_NONE on success.
 *
 * @error
 *   <table class="ec" align="center">
 *     <tr>
 *       <td class="ecl">CPL_ERROR_NULL_INPUT</td>
 *       <td class="ecr">
 *         An input pointer is <tt>NULL</tt>.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td class="ecl">CPL_ERROR_DATA_NOT_FOUND</td>
 *       <td class="ecr">
 *         The input <i>framelist</i> does not contain any input file.
 *         A call to the function cpl_error_get_where() would return
 *         the string "cpl_dfs_setup_product_header/set-of-frames".
 *       </td>
 *     </tr>
 *     <tr>
 *       <td class="ecl">CPL_ERROR_FILE_NOT_FOUND</td>
 *       <td class="ecr">
 *         Any file listed in the input set-of-frames doesn't exist.
 *       </td>
 *     </tr>
 *     <tr>
 *       <td class="ecl">CPL_ERROR_UNSPECIFIED</td>
 *       <td class="ecr">
 *         This error would be caused by any unexpected incompatibility
 *         within the CPL, and it should be immediately reported to the
 *         CPL Team.
 *       </td>
 *     </tr>
 *   </table>
 * @enderror
 *
 * This function checks the @em header associated to a pipeline product,
 * to ensure that it is DICB compliant. In particular, this function does
 * the following:
 *
 *    -# Copy to @em header, if they are present, the following primary
 *       FITS keywords from the first input frame in the @em framelist:
 *       ORIGIN, TELESCOPE, INSTRUME, OBJECT, RA, DEC, EPOCH, EQUINOX,
 *       RADECSYS, DATE-OBS, MJD-OBS, UTC, LST, PI-COI, OBSERVER. If those
 *       keywords are already present in the @em header property list, they
 *       are overwritten only in case they have the same type. If any of
 *       these keywords are present with an unexpected type, a warning is
 *       issued, but the keywords are copied anyway (provided that the
 *       above conditions are fulfilled), and no error is set.
 *
 *    -# Copy all the HIERARCH.ESO._ keywords from the primary FITS header
 *       of the first input frame in @em framelist, with the exception of
 *       the HIERARCH.ESO.DPR._, and of the .PRO._ and .DRS._ keywords if
 *       the first input frame is a calibration. If those keywords are
 *       already present in @em header, they are overwritten.
 *
 *    -# If found, remove the HIERARCH.ESO.DPR._ keywords from @em header.
 *
 *    -# If found, remove the ARCFILE and ORIGFILE keywords from @em header.
 *
 *    -# Add to @em header the following mandatory keywords from the PRO
 *       dictionary: PIPEFILE, PRO.DID, PRO.REC1.ID, PRO.REC1.DRS.ID,
 *       PRO.REC1.PIPE.ID, PRO.CATG, and PRO.TYPE. If those keywords are
 *       already present in @em header, they are overwritten. The keyword
 *       PRO.TYPE is always given the value "REDUCED".
 *
 *    -# Check the existence of the keyword PRO.DATANCOM in @em header. If
 *       this keyword is missing, one is added, with the value of the total
 *       number of raw input frames.
 *
 *    -# Add to @em header the keywords PRO.REC1.RAW1.NAME, PRO.REC1.RAW1.CATG,
 *       PRO.REC1.CAL1.NAME, PRO.REC1.CAL1.CATG, to describe the content of
 *       the input set-of-frames.
 *
 * See the DICB PRO dictionary to have details on the mentioned PRO keywords.
 *
 * @note
 *   The first input frame in @em framelist is the first raw frame in
 *   @em framelist. If there are no raw frames in input, then the first
 *   calibration frame is taken. Non-FITS files are handled as files
 *   with an empty FITS header.
 */

cpl_error_code cpl_dfs_setup_product_header(cpl_propertylist *header,
                                            const cpl_frame *product_frame,
                                            const cpl_frameset *framelist,
                                            const cpl_parameterlist *parlist,
                                            const char *recid,
                                            const char *pipeline_id,
                                            const char *dictionary_id)
{

    char              cval[FITS_LINESZ];
    const char       *datamd5;
    cpl_frame        *frame;
    cpl_frame        *first_frame = NULL;
    cpl_propertylist *plist;
    const cpl_parameter *param;
    int               nraw;
    int               ncal;
    int               npar;
    int               i;

    const char       *kname;

    char              pval[MAX_PLENGTH + 1];


    /*
     * Here is the list of mandatory keywords, first in input header,
     * second in output _image_ header:
     */

    typedef struct {
        const char *key;
        cpl_type type;
    } klist;

    klist mandatory[] =  {{"ORIGIN", CPL_TYPE_STRING},
                          {"TELESCOP", CPL_TYPE_STRING},
                          {"INSTRUME", CPL_TYPE_STRING},
                          {"OBJECT", CPL_TYPE_STRING},
                          {"RA", CPL_TYPE_DOUBLE},
                          {"DEC", CPL_TYPE_DOUBLE},
                          {"EPOCH", CPL_TYPE_STRING},
                          {"EQUINOX", CPL_TYPE_DOUBLE},
                          {"RADECSYS", CPL_TYPE_STRING},
                          {"DATE-OBS", CPL_TYPE_STRING},
                          {"MJD-OBS", CPL_TYPE_DOUBLE},
                          {"UTC", CPL_TYPE_DOUBLE},
                          {"LST", CPL_TYPE_DOUBLE},
                          {"PI-COI", CPL_TYPE_STRING},
                          {"OBSERVER", CPL_TYPE_STRING}};

    const int count_mandatory = (int)(sizeof(mandatory) / sizeof(klist));

    /* Here are the same keywords - as a regular expression */
#define REGMANDATORY "^(ORIGIN|TELESCOP|INSTRUME|OBJECT|RA|DEC|EPOCH|EQUINOX" \
    "|RADECSYS|DATE-OBS|MJD-OBS|UTC|LST|PI-COI|OBSERVER)$"


    cpl_ensure_code(header        != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(product_frame != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(framelist     != NULL, CPL_ERROR_NULL_INPUT);
    cpl_ensure_code(recid         != NULL, CPL_ERROR_NULL_INPUT);

    /* FIXME: 
    cpl_ensure_code(cpl_frame_get_group(product_frame)
                    == CPL_FRAME_GROUP_PRODUCT, CPL_ERROR_ILLEGAL_INPUT);
    */

    /*
     * Forbidden keywords are removed from product header.
     */

    cpl_propertylist_erase_regexp(header, "^ESO DPR |^ARCFILE$|^ORIGFILE$",
                                  0);

    /*
     * Get the first input frame in the input set-of-frames.
     */

    frame = cpl_frameset_get_first(framelist);

    while (frame != NULL) {
        if (cpl_frame_get_group(frame) == CPL_FRAME_GROUP_RAW) {
            first_frame = frame;
            break;
        }
        frame = cpl_frameset_get_next(framelist);
    }

    if (first_frame == NULL) {
        frame = cpl_frameset_get_first(framelist);

        while (frame != NULL) {
            if (cpl_frame_get_group(frame) == CPL_FRAME_GROUP_CALIB) {
                first_frame = frame;
                break;
            }
            frame = cpl_frameset_get_next(framelist);
        }
    }

    if (first_frame == NULL) {
        /* FIXME: Non-standard use of error system */
        char xfid[CPL_MAX_FUNCTION_NAME];

        snprintf(xfid, CPL_MAX_FUNCTION_NAME, "%s/%s", cpl_func,
                 "set-of-frames");

        return cpl_error_set(xfid, CPL_ERROR_DATA_NOT_FOUND);
    }


    /*
     * Now copy all the required entries, if present, from input to
     * product header.
     */

    switch (qfits_is_fits(cpl_frame_get_filename(first_frame))) {
    case -1:
        return cpl_error_set(cpl_func, CPL_ERROR_FILE_NOT_FOUND);
    case 0:
        plist = cpl_propertylist_new();
        break;
    case 1:
        /* Load only the mandatory keywords and the HIERACH ESO keywords */
        plist = cpl_propertylist_load_regexp(cpl_frame_get_filename(first_frame),
                                             0, REGMANDATORY "|^ESO ", 0);
        cpl_ensure_code(plist != NULL, cpl_error_get_code());
        break;
    default:
        cpl_ensure_code(0, CPL_ERROR_UNSPECIFIED);
    }

    for (i = 0; i < count_mandatory; i++) {
        if (cpl_propertylist_contains(plist, mandatory[i].key)) {
            if (cpl_propertylist_get_type(plist, mandatory[i].key) !=
                mandatory[i].type) {
                cpl_msg_warning(cpl_func, "Unexpected type for keyword %s in "
                                "input header", mandatory[i].key);
            }
            if (cpl_propertylist_contains(header, mandatory[i].key)) {
                if (cpl_propertylist_get_type(header, mandatory[i].key) !=
                    mandatory[i].type) {
                    cpl_msg_warning(cpl_func, "Unexpected type for keyword %s "
                                    "in output header", mandatory[i].key);
                }
                if (cpl_propertylist_get_type(plist, mandatory[i].key) !=
                    cpl_propertylist_get_type(header, mandatory[i].key)) {
                    /* Do not copy this keyword */
                    cpl_propertylist_erase(plist, mandatory[i].key);
                }
            }
        }
    }

    /*
     * Here copy all the HIERARCH.ESO._ keywords, excluding the
     * HIERARCH.ESO.DPR._ keywords, and of the .PRO._ and .DRS._
     * keywords.
     */

    /* Also copy the mandatory keywords, if any */

    cpl_propertylist_copy_property_regexp(header, plist,
                                          "^ESO (DPR|DRS|PRO) ", 1);

    cpl_propertylist_delete(plist);


    /*
     * DATAMD5 (placeholder: will be computed by esorex)
     */

    cpl_propertylist_update_string(header, "DATAMD5", "Not computed");
    cpl_propertylist_set_comment(header, "DATAMD5", "MD5 checksum");


    /*
     * PIPEFILE
     */

    kname = cpl_frame_get_filename(product_frame);
    cpl_ensure_code(kname != NULL, cpl_error_get_code());
    cpl_propertylist_update_string(header, PIPEFILE, kname);
    cpl_propertylist_set_comment(header, PIPEFILE, "Filename of data product");


    /*
     * PRO DID
     */

    cpl_propertylist_update_string(header, PRO_DID, dictionary_id);
    cpl_propertylist_set_comment(header, PRO_DID, "Data dictionary for PRO");


    /*
     * PRO CATG
     */

    kname = cpl_frame_get_tag(product_frame);
    cpl_ensure_code(kname != NULL, cpl_error_get_code());
    cpl_propertylist_update_string(header, PRO_CATG, kname);
    cpl_propertylist_set_comment(header, PRO_CATG,
                                 "Category of pipeline product frame");


    /*
     * PRO TYPE
     */

    cpl_propertylist_update_string(header, PRO_TYPE, "REDUCED");
    cpl_propertylist_set_comment(header, PRO_TYPE, "Product type");


    /*
     * PRO REC1 ID
     */

    cpl_propertylist_update_string(header, PRO_REC_ID, recid);
    cpl_propertylist_set_comment(header, PRO_REC_ID,
                                 "Pipeline recipe (unique) identifier");


    /*
     * PRO REC1 DRS ID
     */

    /* cpl_propertylist_update_string(header,
                                      PRO_REC_DRS_ID, cpl_get_version()); */

    snprintf(cval, FITS_LINESZ, "%s-%s", PACKAGE, PACKAGE_VERSION);
    cpl_propertylist_update_string(header, PRO_REC_DRS_ID, cval);
    cpl_propertylist_set_comment(header, PRO_REC_DRS_ID,
                                 "Data Reduction System identifier");


    /*
     * PRO REC1 PIPE ID
     */

/*    snprintf(cval, FITS_LINESZ, "%s/%s", PACKAGE, PACKAGE_VERSION); */
    cpl_propertylist_update_string(header, PRO_REC_PIPE_ID, pipeline_id);
    cpl_propertylist_set_comment(header, PRO_REC_PIPE_ID,
                                 "Pipeline (unique) identifier");


    /*
     * PRO REC1 RAWi NAME  and  PRO REC1 RAWi CATG
     */

    nraw = 0;
    frame = cpl_frameset_get_first(framelist);

    while (frame != NULL) {
        if (cpl_frame_get_group(frame) == CPL_FRAME_GROUP_RAW) {

            ++nraw;

            snprintf(cval, FITS_LINESZ, PRO_REC_RAWi_NAME, nraw);
            cpl_propertylist_update_string(header, cval,
                     qfits_get_base_name(cpl_frame_get_filename(frame)));
            cpl_propertylist_set_comment(header, cval,
                                         "File name of raw frame");

            snprintf(cval, FITS_LINESZ, PRO_REC_RAWi_CATG, nraw);
            cpl_propertylist_update_string(header, cval,
                                           cpl_frame_get_tag(frame));
            cpl_propertylist_set_comment(header, cval,
                                         "Category of raw frame");

        }
        frame = cpl_frameset_get_next(framelist);
    }


    /*
     * PRO DATANCOM
     */

    if (!cpl_propertylist_contains(header, PRO_DATANCOM)) {
        cpl_propertylist_update_int(header, PRO_DATANCOM, nraw);
        cpl_propertylist_set_comment(header, PRO_DATANCOM,
                                     "Number of combined frames");
    }


    /*
     * PRO REC1 CALi NAME,  PRO REC1 CALi CATG,  and  PRO REC1 CALi DATAMD5
     */

    ncal = 0;
    frame = cpl_frameset_get_first(framelist);

    while (frame != NULL) {
        if (cpl_frame_get_group(frame) == CPL_FRAME_GROUP_CALIB) {

            ++ncal;

            snprintf(cval, FITS_LINESZ, PRO_REC_CALi_NAME, ncal);
            cpl_propertylist_update_string(header, cval,
                     qfits_get_base_name(cpl_frame_get_filename(frame)));
            cpl_propertylist_set_comment(header, cval,
                                         "File name of calibration frame");

            snprintf(cval, FITS_LINESZ, PRO_REC_CALi_CATG, ncal);
            cpl_propertylist_update_string(header, cval,
                                           cpl_frame_get_tag(frame));
            cpl_propertylist_set_comment(header, cval,
                                         "Category of calibration frame");

            snprintf(cval, FITS_LINESZ, PRO_REC_CALi_DATAMD5, ncal);

            switch (qfits_is_fits(cpl_frame_get_filename(first_frame))) {
            case -1:
                return cpl_error_set(cpl_func, CPL_ERROR_FILE_NOT_FOUND);
            case 0:
                plist = cpl_propertylist_new();
                break;
            case 1:
                plist =
                    cpl_propertylist_load_regexp(cpl_frame_get_filename(frame),
                                                 0, "^DATAMD5$", 0);
                cpl_ensure_code(plist != NULL, cpl_error_get_code());
                break;
            default:
                cpl_ensure_code(0, CPL_ERROR_UNSPECIFIED);
            }

            if (cpl_propertylist_contains(plist, "DATAMD5")) {
                if (cpl_propertylist_get_type(plist, "DATAMD5") !=
                    CPL_TYPE_STRING) {
                    cpl_msg_warning(cpl_func, "Unexpected type for keyword "
                                    "DATAMD5 in input header");
                }
                else {
                    datamd5 = cpl_propertylist_get_string(plist, "DATAMD5");
                    cpl_propertylist_update_string(header, cval, datamd5);
                    cpl_propertylist_set_comment(header, cval,
                                           "MD5 signature of calib frame");
                }
            }
            cpl_propertylist_delete(plist);
        }
        frame = cpl_frameset_get_next(framelist);
    }

    npar = 0;
    /* The cast is OK, since param is not modified */
    param = cpl_parameterlist_get_first((cpl_parameterlist*)parlist);

    while (param != NULL) {

        ++npar;

        kname = cpl_parameter_get_alias(param, CPL_PARAMETER_MODE_CLI);

        switch (cpl_parameter_get_type(param)) {
        case CPL_TYPE_BOOL:
            if (cpl_parameter_get_bool(param) == 1)
              snprintf(pval, MAX_PLENGTH, "%s", "true");
            else
              snprintf(pval, MAX_PLENGTH, "%s", "false");
            break;
        case CPL_TYPE_INT:
            snprintf(pval, MAX_PLENGTH, "%d", cpl_parameter_get_int(param));
            break;
        case CPL_TYPE_DOUBLE:
            snprintf(pval, MAX_PLENGTH, "%g", cpl_parameter_get_double(param));
            break;
        case CPL_TYPE_STRING:
            snprintf(pval, MAX_PLENGTH, "%s", cpl_parameter_get_string(param));
            break;
        default:

            /*
             * Theoretically impossible to get here
             */

            cpl_ensure_code(0, CPL_ERROR_UNSPECIFIED);
        }

        snprintf(cval, FITS_LINESZ, PRO_REC_PARAMi_NAME, npar);
        cpl_propertylist_update_string(header, cval, kname);

        snprintf(cval, FITS_LINESZ, PRO_REC_PARAMi_VALUE, npar);
        cpl_propertylist_update_string(header, cval, pval);

        /* The cast is OK, since param is not modified */
        param = cpl_parameterlist_get_next((cpl_parameterlist*)parlist);
    }


    return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief  Perform any DFS-compliancy required actions (DATAMD5 update)
  @param  self     The list of frames used by the recipe
  @return CPL_ERROR_NONE or the relevant CPL error code on error

   @error
     <table class="ec" align="center">
       <tr>
         <td class="ecl">CPL_ERROR_NULL_INPUT</td>
         <td class="ecr">
            An input pointer is <tt>NULL</tt>.
         </td>
       </tr>
       <tr>
         <td class="ecl">CPL_ERROR_DATA_NOT_FOUND</td>
         <td class="ecr">
           The input <i>framelist</i> contains a frame of type product with a
           missing filename.
         </td>
       </tr>
       <tr>
         <td class="ecl">CPL_ERROR_BAD_FILE_FORMAT</td>
         <td class="ecr">
           The input <i>framelist</i> contains a frame of type product without
           a FITS card with key 'DATAMD5'.
         </td>
       </tr>
       <tr>
         <td class="ecl">CPL_ERROR_FILE_IO</td>
         <td class="ecr">
           The input <i>framelist</i> contains a frame of type product for which
           the FITS card with key 'DATAMD5' could not be updated.
         </td>
       </tr>
     </table>
   @enderror

 */
/*----------------------------------------------------------------------------*/
cpl_error_code cpl_dfs_update_product_header(cpl_frameset * self)
{

    const char      * fitskey  = "DATAMD5";
    const cpl_frame * frame;
    int               nupdate = 0;


    cpl_ensure_code(self != NULL, CPL_ERROR_NULL_INPUT);

    for (frame = cpl_frameset_get_first(self);
         frame != NULL;
         frame = cpl_frameset_get_next(self)) {

        if (cpl_frame_get_group(frame) == CPL_FRAME_GROUP_PRODUCT) {

            const cpl_frame_type type = cpl_frame_get_type(frame);

            if (type == CPL_FRAME_TYPE_TABLE || type == CPL_FRAME_TYPE_IMAGE) {

                const char * filename;
                const char * md5sum;
                char       * newcard;
                int error;


                filename = cpl_frame_get_filename(frame);

                cpl_ensure_code(filename != NULL, CPL_ERROR_DATA_NOT_FOUND);

                md5sum   = qfits_datamd5(filename);

                cpl_ensure_code(md5sum != NULL, CPL_ERROR_BAD_FILE_FORMAT);


                newcard = cpl_sprintf("%s = '%s' / MD5 checksum", fitskey,
                                      md5sum);

                /* This should not be able to fail */
                cpl_ensure_code(newcard != NULL, cpl_error_get_code());

                error = qfits_replace_card(filename, fitskey, newcard);

                cpl_free(newcard);

                cpl_ensure_code(!error, CPL_ERROR_FILE_IO);

                nupdate++;
            }
        }
    }

    cpl_msg_debug(cpl_func, "Updated the FITS card with key '%s' in the primary "
                  "FITS header of %d product(s)", fitskey, nupdate);

    return CPL_ERROR_NONE;

}

/**@}*/
