/* $Id: cpl_frameset_io.c,v 1.7 2006/03/08 15:00:25 llundin Exp $
 *
 * This file is part of the ESO Common Pipeline Library
 * Copyright (C) 2001-2004 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/*
 * $Author: llundin $
 * $Date: 2006/03/08 15:00:25 $
 * $Revision: 1.7 $
 * $Name: cpl-3_1 $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <string.h>

#include <qfits.h>

#include "cpl_frameset_io.h"

#include "cpl_imagelist_defs.h"


/**
 * @defgroup cpl_frameset_io    Frame Sets IO functions
 *
 * @par Synopsis:
 * @code
 *   #include <cpl_frameset_io.h>
 * @endcode
 */

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Load an imagelist from a frameset
  @param    fset        The frames set
  @param    itype       The required image type
  @param    pnum        The plane number 
  @param    extnum      The required extension 
  @return   the loaded list of images or NULL in error case.
  
  extnum = -1 is for all extensions.
  extnum = 0 is for the primary unit.
  extnum = n in for extension number n.

  pnum = 0 is for all planes.
  pnum = n is for plane number n.

  The returned cpl_imagelist must be deallocated using cpl_imagelist_delete()

  Possible #_cpl_error_code_ set in this function:
  - CPL_ERROR_NULL_INPUT if fset is NULL
  - CPL_ERROR_ILLEGAL_INPUT if pnum is negative, extnum is lower than -1
        or one image cannot be loaded as specified
 */
/*----------------------------------------------------------------------------*/
cpl_imagelist * cpl_imagelist_load_frameset(
        const cpl_frameset  *   fset,
        cpl_type                itype,
        int                     pnum, 
        int                     extnum) 
{
    cpl_imagelist   *   imlist ;
    int                 nframes ;
    cpl_frame       *   cur_frame ;
    const char      *   cur_fname ;
    cpl_image       *   img ;
    int                 nb_images_ext ;
    int                 nb_extensions ;
    int                 curr_im ;
    int                 i, j, k ;

    /* Test entries */
    cpl_ensure(fset, CPL_ERROR_NULL_INPUT, NULL) ;
    cpl_ensure(pnum>=0 && extnum >=-1, CPL_ERROR_ILLEGAL_INPUT, NULL) ;

    /* Number of frames */
    nframes = cpl_frameset_get_size(fset) ;

    /* Create output image list for the moment */
    imlist = cpl_imagelist_new() ;

    /* Loop on the images to load them */
    curr_im = 0 ;
    for (k=0 ; k<nframes ; k++) {
        /* Get the current frame name */
        cur_frame = cpl_frameset_get_frame(fset, k) ;
        cur_fname = cpl_frame_get_filename(cur_frame) ;
        /* Get the nb of extensions */
        nb_extensions = qfits_query_n_ext((char*)cur_fname) ;
        if (extnum < 0) {
            /* All extensions required */
            for (i=0 ; i<=nb_extensions ; i++) {
                /* Count the number of images in the extension */
                nb_images_ext = qfits_query_nplanes((char*)cur_fname, i) ;
                if (pnum == 0) {
                    /* All planes required */
                    for (j=1 ; j<=nb_images_ext ; j++) {
                        img = cpl_image_load(cur_fname, itype, j-1, i);
                        if (cpl_imagelist_set(imlist, img, curr_im) !=
                                CPL_ERROR_NONE) {
                            cpl_imagelist_delete(imlist) ;
                            cpl_image_delete(img) ;
                            cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL);
                        }
                        curr_im++ ;
                    }
                } else {
                    /* Only one plane required */
                    img = cpl_image_load(cur_fname, itype, pnum-1, i);
                    if (cpl_imagelist_set(imlist, img, curr_im) !=
                            CPL_ERROR_NONE) {
                        cpl_imagelist_delete(imlist) ;
                        cpl_image_delete(img) ;
                        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL);
                    }
                    curr_im++ ;
                }
            }
        } else {
            /* Only extension nb extnum required */
            nb_images_ext = qfits_query_nplanes((char*)cur_fname, extnum) ;
            if (pnum == 0) {
                /* All planes required */
                for (j=1 ; j<=nb_images_ext ; j++) {
                    img = cpl_image_load(cur_fname, itype, j-1, extnum);
                    if (cpl_imagelist_set(imlist, img, curr_im) !=
                            CPL_ERROR_NONE) {
                        cpl_imagelist_delete(imlist) ;
                        cpl_image_delete(img) ;
                        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL);
                    }
                    curr_im++ ;
                }
            } else {
                /* Only one plane required */
                img = cpl_image_load(cur_fname, itype, pnum-1, extnum);
                if (cpl_imagelist_set(imlist, img, curr_im) !=
                        CPL_ERROR_NONE) {
                    cpl_imagelist_delete(imlist) ;
                    cpl_image_delete(img) ;
                    cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL);
                }
                curr_im++ ;
            }
        }
    }

    /* Check if the create image set is correct */
    if (cpl_imagelist_is_uniform(imlist) != 0) {
        cpl_imagelist_delete(imlist) ;
        cpl_ensure(0, CPL_ERROR_ILLEGAL_INPUT, NULL) ;
    }
    return imlist ;
}

/**@}*/
